﻿using System.Collections.ObjectModel;
using Infragistics.Olap.Data;
using System.Collections.Generic;
using Infragistics.Controls.Grids;
using System;
using Infragistics.Olap;
using System.Threading;
using Infragistics.Olap.Data.Base;
using System.Runtime.Serialization;
using System.Collections;

namespace SaveLoadSerializationSample
{
    public class SavedPivotGridView
    {
        #region SavedInformation

        private string timestamp;
        public string Timestamp
        {
            get { return timestamp; }
            set { timestamp = value; }
        }

        public int SelectedCubeIndex;

        //unique names of each row, column, filter and measure that will be saved
        public List<string> RowHierarchies;
        public List<string> ColumnHierarchies;
        public List<string> FilterHierarchies;
        public List<string> Measures;

        //unique name of each member to save and its IsExpanded property
        public List<CustomDictItem> FilterMembersToExpand;
        //unique name of each member to save and its IsSelected property
        public List<CustomDictItem> FilterMembersToLoad;

        #endregion SavedInformation

        #region SaveMethods

        public void SaveView(XamPivotGrid pivotGrid)
        {
            //check if pivotGrid actually has information that can be saved
            if (pivotGrid.DataSource == null)
                return;

            if (pivotGrid.DataSource.Cube == null)
                return;

            //give a name for the saved view - default is the time stamp of the save moment
            this.Timestamp = DateTime.Now.ToString();

            //save the selected cube
            this.SelectedCubeIndex = pivotGrid.DataSource.Cubes.IndexOf(pivotGrid.DataSource.Cube);

            //save the row hierarchies
            this.RowHierarchies = new List<string>();
            foreach (IAreaItemViewModel item in pivotGrid.DataSource.Rows)
            {
                var FilterViewModel = item as IFilterViewModel;
                if (FilterViewModel != null)
                {
                    this.RowHierarchies.Add(FilterViewModel.Hierarchy.UniqueName);
                }
            }

            //save the column hierarchies
            this.ColumnHierarchies = new List<string>();
            foreach (IAreaItemViewModel item in pivotGrid.DataSource.Columns)
            {
                var FilterViewModel = item as IFilterViewModel;
                if (FilterViewModel != null)
                {
                    this.ColumnHierarchies.Add(FilterViewModel.Hierarchy.UniqueName);
                }
            }

            //save the filter hierarchies
            this.FilterHierarchies = new List<string>();
            foreach (IAreaItemViewModel item in pivotGrid.DataSource.Filters)
            {
                var FilterViewModel = item as IFilterViewModel;
                if (FilterViewModel != null)
                {
                    this.FilterHierarchies.Add(FilterViewModel.Hierarchy.UniqueName);
                }
            }

            //save the measures
            this.Measures = new List<string>();
            foreach (IAreaItemViewModel item in pivotGrid.DataSource.Measures)
            {
                var measureViewModel = item as IMeasureViewModel;
                if (measureViewModel != null)
                {
                    this.Measures.Add(measureViewModel.Measure.UniqueName);
                }
            }

            //save the filter members expand states. this is needed so we can expand the levels of the tree that contains the filter members.
            //only if we have this info we can go through all filter members in the filter hierarchies and restore their selection states
            FilterMembersToExpand = new List<CustomDictItem>();
            foreach (IFilterViewModel filterViewModel in pivotGrid.DataSource.Filters)
            {
                if (filterViewModel.FilterMembers != null)
                    foreach (IFilterMember filterMember in filterViewModel.FilterMembers)
                        SaveExpansionStateOfFilterMembers(filterMember);
            }
            foreach (PivotHeaderCell cell in pivotGrid.GridLayout.RowHeaderCells)
            {
                CustomDictItem item = new CustomDictItem(cell.Member.UniqueName, cell.IsExpanded);

                if (!CustomDictContains(this.FilterMembersToExpand, item.UniqueName))
                {
                    this.FilterMembersToExpand.Add(item);
                }
                else
                {
                    GetCustomDictItem(this.FilterMembersToExpand, item.UniqueName).Value = true;
                }
            }
            foreach (PivotHeaderCell cell in pivotGrid.GridLayout.ColumnHeaderCells)
            {
                CustomDictItem item = new CustomDictItem(cell.Member.UniqueName, cell.IsExpanded);

                if (!CustomDictContains(this.FilterMembersToExpand, item.UniqueName))
                {
                    this.FilterMembersToExpand.Add(item);
                }
                else
                {
                    GetCustomDictItem(this.FilterMembersToExpand, item.UniqueName).Value = true;
                }
            }

            //save the filter members selected states for column, row and filters hierarchies.
            //this is needed so we can restore the selected filter members from the filter menu of each hierarchy added to the grid
            FilterMembersToLoad = new List<CustomDictItem>();
            for (int i = 0; i < pivotGrid.DataSource.Columns.Count; i++)
            {
                IFilterViewModel filterViewModel = pivotGrid.DataSource.Columns[i] as IFilterViewModel;

                if (filterViewModel != null)
                {
                    if (filterViewModel.FilterMembers != null)
                        foreach (IFilterMember filterMember in filterViewModel.FilterMembers)
                            SaveSelectionStateOfFilterMembers(filterMember);
                }
            }
            for (int i = 0; i < pivotGrid.DataSource.Rows.Count; i++)
            {
                IFilterViewModel filterViewModel = pivotGrid.DataSource.Rows[i] as IFilterViewModel;

                if (filterViewModel != null)
                {
                    if (filterViewModel.FilterMembers != null)
                        foreach (IFilterMember filterMember in filterViewModel.FilterMembers)
                            SaveSelectionStateOfFilterMembers(filterMember);
                }
            } 
            foreach (IFilterViewModel filterViewModel in pivotGrid.DataSource.Filters)
            {
                if (filterViewModel.FilterMembers != null)
                    foreach (IFilterMember filterMember in filterViewModel.FilterMembers)
                        SaveSelectionStateOfFilterMembers(filterMember);
            }
        }

        //recusive method to save the expansion states of the filter members
        private void SaveExpansionStateOfFilterMembers(IFilterMember filterMember)
        {
            if (filterMember.FilterMembers != null)
            {
                foreach (IFilterMember subFilterMember in filterMember.FilterMembers)
                {
                    SaveExpansionStateOfFilterMembers(subFilterMember);
                }
            }

            CustomDictItem item = new CustomDictItem(filterMember.Member.UniqueName, filterMember.IsExpanded);

            if (!CustomDictContains(this.FilterMembersToExpand, item.UniqueName))
            {
                this.FilterMembersToExpand.Add(item);
            }
            else
            {
                GetCustomDictItem(this.FilterMembersToExpand, item.UniqueName).Value = true;
            }
        }

        //recusive method to save the selection states of the filter members
        private void SaveSelectionStateOfFilterMembers(IFilterMember filterMember)
        {
            if (filterMember.FilterMembers != null)
            {
                foreach (IFilterMember subFilterMember in filterMember.FilterMembers)
                {
                    SaveSelectionStateOfFilterMembers(subFilterMember);
                }
            }

            if (filterMember.IsSelected.HasValue)
            {
                CustomDictItem item = new CustomDictItem(filterMember.Member.UniqueName, filterMember.IsSelected.Value);

                if (!CustomDictContains(this.FilterMembersToLoad, item.UniqueName))
                {
                    this.FilterMembersToLoad.Add(item);
                }
                else
                {
                    GetCustomDictItem(this.FilterMembersToLoad, item.UniqueName).Value = true;
                }
            }
        }

        #endregion SaveMethods

        #region LoadMethods

        public void Fill(XamPivotGrid pivotGrid)
        {
            //this flag facilitates the timing of the load events
            StateLoadComplete = false;

            //select the stored cube index
            pivotGrid.DataSource.Cube = pivotGrid.DataSource.Cubes[this.SelectedCubeIndex];

            //enable deferred layout update so all restore procedures are executed at once. 
            //otherwise the UI will be updated after each command and thus it will be slow
            pivotGrid.DataSource.DeferredLayoutUpdate = true;

            //clear and then load the saved row, column and filter hierarchies and then the saved measures
            pivotGrid.DataSource.Rows.Clear();
            pivotGrid.DataSource.Columns.Clear();
            pivotGrid.DataSource.Measures.Clear();
            pivotGrid.DataSource.Filters.Clear();

            foreach (IDimension dim in pivotGrid.DataSource.Cube.Dimensions)
            {
                foreach (IHierarchy hierarchy in dim.Hierarchies)
                {
                    if (this.RowHierarchies.Contains(hierarchy.UniqueName))
                    {
                        IFilterViewModel fvm = pivotGrid.DataSource.CreateFilterViewModel(hierarchy);
                        pivotGrid.DataSource.Rows.Add(fvm);
                    }
                    else if (this.ColumnHierarchies.Contains(hierarchy.UniqueName))
                    {
                        IFilterViewModel fvm = pivotGrid.DataSource.CreateFilterViewModel(hierarchy);
                        pivotGrid.DataSource.Columns.Add(fvm);
                    }
                    else if (this.FilterHierarchies.Contains(hierarchy.UniqueName))
                    {
                        IFilterViewModel fvm = pivotGrid.DataSource.CreateFilterViewModel(hierarchy);
                        pivotGrid.DataSource.Filters.Add(fvm);
                    }
                }
            }
            foreach (IMeasure measure in pivotGrid.DataSource.Cube.Measures)
            {
                if (this.Measures.Contains(measure.UniqueName))
                {
                    IMeasureViewModel mvm = pivotGrid.DataSource.CreateMeasureViewModel(measure);
                    pivotGrid.DataSource.Measures.Add(mvm);
                }
            }

            //attach to the layoutloaded event so expansion of all hierarchy levels can be executed after the UI is updated
            if (!eventAttached)
                pivotGrid.LayoutLoaded += pivotGrid_LayoutLoaded;

            //turn of the deferred layout update and force an UI udpate
            pivotGrid.DataSource.DeferredLayoutUpdate = false;
        }

        void pivotGrid_LayoutLoaded(object sender, EventArgs e)
        {
            if (!this.StateLoadComplete)
            {
                StateLoadComplete = true;

                //use deferred updates to execute all operations at once
                (sender as XamPivotGrid).DataSource.DeferredLayoutUpdate = true;

                //expand members for columns, rows and filters
                for (int i = 0; i < (sender as XamPivotGrid).DataSource.Rows.Count; i++)
                {
                    IFilterViewModel filterViewModel = (sender as XamPivotGrid).DataSource.Rows[i] as IFilterViewModel;

                    if (filterViewModel != null)
                    {
                        if (filterViewModel.FilterMembers != null)
                            foreach (IFilterMember filterMember in filterViewModel.FilterMembers)
                                ExpandMembers(filterViewModel, filterMember);
                    }
                }
                for (int i = 0; i < (sender as XamPivotGrid).DataSource.Columns.Count; i++)
                {
                    IFilterViewModel filterViewModel = (sender as XamPivotGrid).DataSource.Columns[i] as IFilterViewModel;

                    if (filterViewModel != null)
                    {
                        if (filterViewModel.FilterMembers != null)
                            foreach (IFilterMember filterMember in filterViewModel.FilterMembers)
                                ExpandMembers(filterViewModel, filterMember);
                    }
                }
                foreach (IFilterViewModel filterViewModel in (sender as XamPivotGrid).DataSource.Filters)
                {
                    if (filterViewModel.FilterMembers != null)
                        foreach (IFilterMember filterMember in filterViewModel.FilterMembers)
                            ExpandMembers(filterViewModel, filterMember);
                }

                //select members for columns, rows and filters
                foreach (IFilterViewModel filterViewModel in (sender as XamPivotGrid).DataSource.Filters)
                {
                    if (filterViewModel.FilterMembers != null)
                        foreach (IFilterMember filterMember in filterViewModel.FilterMembers)
                            SelectMembers(filterMember);
                }
                for (int i = 0; i < (sender as XamPivotGrid).DataSource.Rows.Count; i++)
                {
                    IFilterViewModel filterViewModel = (sender as XamPivotGrid).DataSource.Rows[i] as IFilterViewModel;

                    if (filterViewModel != null)
                    {
                        if (filterViewModel.FilterMembers != null)
                            foreach (IFilterMember filterMember in filterViewModel.FilterMembers)
                                SelectMembers(filterMember);
                    }
                }
                for (int i = 0; i < (sender as XamPivotGrid).DataSource.Columns.Count; i++)
                {
                    IFilterViewModel filterViewModel = (sender as XamPivotGrid).DataSource.Columns[i] as IFilterViewModel;

                    if (filterViewModel != null)
                    {
                        if (filterViewModel.FilterMembers != null)
                            foreach (IFilterMember filterMember in filterViewModel.FilterMembers)
                                SelectMembers(filterMember);
                    }
                }

                //stop the deferred update and force a refresh of the UI
                (sender as XamPivotGrid).DataSource.DeferredLayoutUpdate = false;
            }
        }

        private void SelectMembers(IFilterMember filterMember)
        {
            if (filterMember.FilterMembers != null)
            {
                foreach (IFilterMember subFilterMember in filterMember.FilterMembers)
                {
                    SelectMembers(subFilterMember);
                }
            }

            if (CustomDictContains(FilterMembersToLoad, filterMember.Member.UniqueName))
                filterMember.IsSelected = GetCustomDictItem(FilterMembersToLoad, filterMember.Member.UniqueName).Value;
        }

        private void ExpandMembers(IFilterViewModel filterViewModel, IFilterMember filterMember)
        {
            //use this event for syncronizing the otherwise async load of members
            ManualResetEvent manualResetEvent = new ManualResetEvent(false);
            filterViewModel.LoadFilterMembersCompleted += (sender1, e1) => manualResetEvent.Set();

            if (CustomDictContains(FilterMembersToExpand, filterMember.Member.UniqueName))
                filterMember.IsExpanded = GetCustomDictItem(FilterMembersToExpand, filterMember.Member.UniqueName).Value;

            if (!filterMember.IsExpanded)
                return;

            manualResetEvent.WaitOne();

            foreach (IFilterMember fm in filterMember.FilterMembers)
            {
                ExpandMembers(filterViewModel, fm);
            }

        }

        #endregion LoadMethods

        #region Helper fields and methods

        private bool StateLoadComplete;
        private bool eventAttached = false;

        private CustomDictItem GetCustomDictItem(List<CustomDictItem> ListOfItems, string UniqueName)
        {
            foreach (CustomDictItem item in ListOfItems)
            {
                if (item.UniqueName == UniqueName)
                    return item;
            }
            throw new Exception("Dict does not contain this item");
        }

        private bool CustomDictContains(List<CustomDictItem> ListOfItems, string UniqueName)
        {
            foreach (CustomDictItem item in ListOfItems)
            {
                if (item.UniqueName == UniqueName)
                    return true;
            }
            return false;
        }


        #endregion Helper fields and methods
    }

    //class to store UniqueNames and the bool values for the corresponding member
    public class CustomDictItem
    {
        public string UniqueName;
        public bool Value;

        //a parameterless constructor is needed in order to make this class serializable
        public CustomDictItem()
        {

        }

        public CustomDictItem(string UniqueName, bool Value)
        {
            this.UniqueName = UniqueName;
            this.Value = Value;
        }
    }
}
